/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.2
 * 
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: event.c,v $
 *
 * HISTORY
 * $Log: event.c,v $
 * Revision 1.7  1994/06/11  00:02:24  menze
 * Added copyright notice
 *
 * Revision 1.6  1994/05/07  17:28:09  davidm
 * (fire): event handlers now get unique names when XK_DEBUG is on.
 *
 * Revision 1.5  1994/04/22  02:07:58  davidm
 * Soft-clock now gets invoked only every other SCOUT_TICKS_PER_CLOCK
 * ticks.  Corrected for that.
 *
 * Revision 1.4  1994/04/22  01:47:49  davidm
 * (evInit): Fixed race condition (scout_clock_thread is volatile!).
 *
 * Revision 1.3  1994/03/30  00:55:24  davidm
 * (softClock): Modified to use a continuation instead of a while loop and
 * a regular thread-block.
 *
 * Revision 1.2  1994/03/30  00:34:02  davidm
 * Adapted to Scout.
 *
 * Revision 1.1  1994/03/12  01:40:45  davidm
 * Initial revision
 */
/*
 * Event manager.  This module implements a facility for managing
 * timeouts.  Both, scheduling timeouts via evSchedule() and canceling
 * them via evCancel() are as efficient as posible.  This is achieved
 * via a "timing wheel."  It consists of a number (XK_EVENT_WHEEL_SIZE) of
 * queues.  All events with a firing time that have the same low-order
 * bits (modulo XK_EVENT_WHEEL_SIZE) go into the same queue.  With an
 * appropriate choice of XK_EVENT_WHEEL_SIZE, each queue therefore will be
 * relatively short, guaranteeing efficient insertion into a queue.
 * Furthermore, with each tick, the clock routine has to inspect a
 * single queue only to find events that are ready to fire.
 */
#include "scout_config.h"
#include "scout_thread.h"
#include "scout_time.h"
#include "event.h"
#include "xk_assert.h"
#include "xk_debug.h"
#include "platform.h"
#include "event_i.h"

int traceevent = 0;

static u_long		softTime = 0;	/* softClock's idea of "now" */
static int		usecPerTick;	/* interval between ticks */
static struct Event	wheel[XK_EVENT_WHEEL_SIZE];
static Scout_Thread_Options options;
static int		nThreadsCached;
static Scout_Thread	*threadCache;
static int		nEventsCached;
static Event		eventCache;


#ifdef XK_THREAD_TRACE

Map		localEventMap;
static Map	threadEventMap;

static Event	self (void);

#endif /* XK_THREAD_TRACE */

static void start(void *arg);	/* forward declaration */

/*
 * Tell GCC that cacheThread never returns:
 */
static void cacheThread (void) XNORETURN;


static void
cacheEvent(Event e)
{
#ifdef XK_THREAD_TRACE
    xAssert(e);
    mapUnbind(localEventMap, &e);
#endif /* XK_THREAD_TRACE */

    if (nEventsCached < XK_EVENT_EVENT_CACHE_SIZE) {
	xTrace1(event, TR_DETAILED, "cacheEvent: caching event 0x%p", e);
	++nEventsCached;
	e->next = eventCache;
	eventCache = e;
    } else {
	xTrace1(event, TR_DETAILED, "cacheEvent: freeing event 0x%p", e);
	xFree(e);
    } /* if */
} /* cacheEvent */


static Event
getEvent(void)
{
    Event e;

    if (nEventsCached > 0) {
	--nEventsCached;
	e = eventCache;
	eventCache = eventCache->next;
	xTrace1(event, TR_DETAILED,
		"getEvent: picked up event 0x%p from cache", e);
    } else {
	/* no cached event---have to create a new one */
	e = xMalloc(sizeof(struct Event));
	xTrace1(event, TR_DETAILED, "getEvent: allocated event 0x%p", e);
    } /*if */
    xAssert(e);
    return e;
} /* getEvent */


static void
cacheThread(void)
{
    if (nThreadsCached < XK_EVENT_THREAD_CACHE_SIZE) {
	xTrace1(event, TR_DETAILED, "cacheThread: caching thread 0x%p",
		scout_thread_self());
	++nThreadsCached;
	scout_thread_self()->next = threadCache;
	threadCache = scout_thread_self();
	scout_thread_suspend_with_continuation();
    } else {
	xTrace1(event, TR_DETAILED, "cacheThread: stopping thread 0x%p",
		scout_thread_self());
	scout_thread_stop();
    } /* if */
} /* cacheThread */


static void
fire(Event e)
{
    Scout_Thread *t;

    xTrace1(event, TR_FUNCTIONAL_TRACE, "fire(event=0x%p)", e);
    if (nThreadsCached > 0) {
	--nThreadsCached;
	t = threadCache;
	threadCache = threadCache->next;
	scout_thread_set_cont_arg(t, e);
	xTrace1(event, TR_DETAILED,
		"fire: picked up thread 0x%p from cache", t);
    } else {
#ifdef XK_DEBUG
	static int nhandlers = 0;
	char buf[32];

	sprintf(buf, "event-handler-%d", ++nhandlers);
	options.name = buf;
#endif
	options.arg = e;
	t = scout_thread_create(start, &options);
	xTrace1(event, TR_DETAILED, "fire: created thread 0x%p", t);
    } /* if */
    xAssert(t);
    e->state = E_SCHEDULED;
#ifdef XK_THREAD_TRACE
    xGetTime(&e->startTime);
#endif /* XK_THREAD_TRACE */
    scout_thread_wakeup(t);
} /* fire */


/*
 * Enqueue "e" into a circular doubly linked list after element
 * "after".
 */
static void
enqueue(Event e, Event after)
{
    e->prev = after;
    e->next = after->next;
    after->next = e;
    e->next->prev = e;
} /* enqueue */


/*
 * Dequeue "e" from a circular doubly linked list.  It is guaranted
 * that the queue will remain non-empty.
 */
static inline void
dequeue(Event e)
{
    e->prev->next = e->next;
    e->next->prev = e->prev;
} /* dequeue */


/*
 * When going from time "now" to "now+1", all events that should fire
 * in the interval (now-1*usecPerTick,now*usecPerTick] are fired.
 */
static void
softClock(void *arg)
{
    u_long time, newTime;
    Event e, q;

    /* atomically read current time: */
    newTime = scout_age/SCOUT_TICKS_PER_CLOCK;

    xTrace1(event, TR_DETAILED, "softClock: %lu ticks",
	    newTime - softTime);

    for (time = softTime; time != newTime; ++time) {
	q = &wheel[time % XK_EVENT_WHEEL_SIZE];
	for (e = q->next; e != q && e->deltat == 0; e = e->next) {
	    dequeue(e);
	    fire(e);
	} /* for */
	--e->deltat;
    } /* for */

    softTime = newTime;
    /*
     * Put ourselves to rest.  Notice that it is OK for
     * hardClock() to wake us up before we suspend.
     */
    xAssert(scout_thread_self());
    scout_clock_thread = scout_thread_self();
    scout_thread_suspend_with_continuation();
} /* softClock */


/*
 * Insert event "newEvent" into "q".  The queue is maintained in a
 * sorted order such that the "deltat" value of events in the queue
 * specifies how many revolutions after the *preceeding* event the
 * event fires.
 */
static void
insert(Event newEvent, Event head)
{
    Event curr;

    xTrace2(event, TR_FUNCTIONAL_TRACE, "insert(newEvent=0x%p,head=0x%p)",
	    newEvent, head);

    for (curr = head->next; curr != head; curr = curr->next) {
	if (curr->deltat < newEvent->deltat) {
	    /* newEvent goes after curr: */
	    newEvent->deltat -= curr->deltat;
	} else {
	    /* newEvent goes in front of curr: */
	    curr->deltat -= newEvent->deltat;
	    enqueue(newEvent, curr->prev);
	    return;
	} /* if */
    } /* for */
    /* newEvent goes at the end */
    enqueue(newEvent, head->prev);
} /* insert */


/*
 * Fired events start here:
 */
static void
start(void *arg)
{
    Event e = (Event) arg;

    xTrace1(event, TR_FUNCTIONAL_TRACE, "start(e=0x%p)", e);
    xAssert(e->state == E_SCHEDULED);
    if (e->flags & E_CANCELLED_F) {
	xTrace1(event, TR_MORE_EVENTS,
		"start: not starting cancelled event 0x%p", e);
	cacheEvent(e);
	cacheThread();
    } /* if */

#ifdef XK_THREAD_TRACE
    {
	Scout_Thread *t;
	
	xAssert(mapResolve(localEventMap, &e, 0) == XK_SUCCESS);
	t = scout_thread_self();
	e->bind = mapBind(threadEventMap, &t, e);
	xAssert(e->bind != ERR_BIND);
	xGetTime(&e->startTime);
    }
#endif /* XK_THREAD_TRACE */

    e->state = E_RUNNING;
    xTrace1(event, TR_EVENTS, "start: invoking event-handler at 0x%p",
	    e->func);
    (*e->func)(e, e->arg);
    e->state = E_FINISHED;

#ifdef XK_THREAD_TRACE
    xAssert(e->bind);
    mapRemoveBinding(threadEventMap, e->bind);
    e->bind = 0;
#endif /* XK_THREAD_TRACE */

    if (e->flags & E_DETACHED_F) {
	cacheEvent(e);
    } /* if */
    cacheThread();
} /* start */


void
evInit(int usec)
{
    int i;
    Scout_Thread *t;

    xTrace1(event, TR_FUNCTIONAL_TRACE, "evInit(interval=%d)", usec);

    usecPerTick = usec;

#ifdef XK_THREAD_TRACE
    localEventMap = mapCreate(XK_EVENT_WHEEL_SIZE, sizeof(Event));
    threadEventMap = mapCreate(XK_EVENT_WHEEL_SIZE, sizeof(Scout_Thread*));
#endif /* XK_THREAD_TRACE */

    for (i = 0; i < XK_EVENT_WHEEL_SIZE; ++i) {
	wheel[i].next = &wheel[i];
	wheel[i].prev = &wheel[i];
	wheel[i].flags = 0;
    } /* for */
 
    options = *scout_thread_default_options;
    options.name = "soft-clock";
    options.priority = SCOUT_THREAD_PRIORITY_MAX;
    t = scout_thread_create(softClock, &options);
    if (!t) {
	xAbort("evInit: failed to create soft-clock thread!");
    } /* if */
    xTrace2(event, TR_EVENTS, "evInit: created thread `%s' at 0x%p",
	    options.name, t);
    scout_clock_thread = t;

    options = *scout_thread_default_options;
    options.name = "event-handler";

    xTrace0(event, TR_FUNCTIONAL_TRACE, "evInit: exiting");
} /* evInit */


Event
evSchedule(EvFunc fun, void *arg, unsigned usec)
{
    Event e;
    int nTicks;

    xTrace3(event, TR_FUNCTIONAL_TRACE,
	    "evSchedule(fun=0x%p,arg=0x%p,usec=%u)", fun, arg, usec);

    e = getEvent();
    e->func = fun;
    e->arg = arg;

#ifdef XK_THREAD_TRACE
    {
	XTime offset;
	Bind b;

	xTrace1(event, TR_MORE_EVENTS, "evSchedule: adding event 0x%p to maps",
		e);
	offset.sec  = usec / (1000 * 1000);
	offset.usec = usec % (1000 * 1000);
	xAssert(localEventMap);
	b = mapBind(localEventMap, &e, e);
	xAssert(b != ERR_BIND);
	e->bind = 0;	/* filled in in 'start' once a thread is assigned */
	xGetTime(&e->startTime);
	xTrace2(event, TR_DETAILED, "event start time: %lu %lu",
		e->startTime.sec, e->startTime.usec);
	xAddTime(&e->startTime, e->startTime, offset);
    }
#endif /* XK_THREAD_TRACE */

    if (usec > 0) {
	/* rounded # of ticks to delay: */
	nTicks = (2*usec + usecPerTick) / (2*usecPerTick);
	if (nTicks == 0) {
	    ++nTicks;		/* delay at least one tick */
	} /* if */
	/*
	 * Notice: softClock can be behind the real time due to other
	 *	   threads not relinquishing the processor frequently
	 *	   enough.  We take care of this by adding the
	 *	   difference between the real time (scout_age) and
	 *	   softClock's time to nTicks:
	 */
	nTicks = (scout_age/SCOUT_TICKS_PER_CLOCK + nTicks) - softTime;

	/* compute how many clock-wheel revolutions this needs: */
	e->deltat = nTicks / XK_EVENT_WHEEL_SIZE;
	e->flags  = 0;
	e->state  = E_PENDING;
	insert(e, &wheel[(softTime + nTicks) % XK_EVENT_WHEEL_SIZE]);
    } else {
	e->flags  = 0;
	e->deltat = 0;
	fire(e);
    } /* if */
    xTrace0(event, TR_FUNCTIONAL_TRACE, "evSchedule: exiting");
    return e;
} /* evSchedule */


void
evDetach(Event e)
{
    xTrace1(event, TR_FUNCTIONAL_TRACE, "evDetach(e=0x%p)", e);
    if (e->state == E_FINISHED) {
	xTrace1(event, TR_EVENTS, "evDetach: caching event 0x%p", e);
	cacheEvent(e);
    } else {
	xTrace1(event, TR_EVENTS, "evDetach: marks event 0x%p as detached", e);
	e->flags |= E_DETACHED_F;
    } /* if */
    xTrace0(event, TR_FUNCTIONAL_TRACE, "evDetach: exiting");
} /* evDetach */


EvCancelReturn
evCancel(Event e)
{
    EvCancelReturn res = EVENT_CANCELLED;

    xTrace1(event, TR_FUNCTIONAL_TRACE, "evCancel(event=0x%p)", e);
    xAssert(e);
    e->flags |= E_DETACHED_F | E_CANCELLED_F;
    switch (e->state) {

      case E_SCHEDULED:
	/* 
	 * "start" will catch this event before it gets a chance to run:
	 */
	xTrace2(event, TR_MORE_EVENTS,
		"evCancel: cancelled event 0x%p in SCHEDULED state, flags %#x",
		e, e->flags);
	break;

      case E_BLOCKED:
      case E_RUNNING:
	res = EVENT_RUNNING;
	break;

      case E_FINISHED:
	res = EVENT_FINISHED;
	cacheEvent(e);
	break;

      case E_PENDING:
	e->next->deltat += e->deltat;
	dequeue(e);
	cacheEvent(e);
	break;

      default:
	xError("evCancel: broken event state");
    } /* switch */
    return res;
} /* evCancel */


int
evIsCancelled(Event e)
{
    xAssert(e);
    return e->flags & E_CANCELLED_F;
} /* evIsCancelled */


/*
 * Debugging support (thread tracing):
 */

#ifdef XK_THREAD_TRACE

static Event
self()
{
    Event e;
    Scout_Thread *t;

    t = scout_thread_self();
    if (mapResolve(threadEventMap, &t, &e) == XK_SUCCESS) {
	return e;
    } else {
	xTrace0(event, TR_EVENTS, "self: couldn't find event structure");
	return 0;
    } /* if */
} /* self */

#endif /* XK_THREAD_TRACE */


void
evCheckStack(char *name)
{
#ifdef XK_THREAD_TRACE
    if (scout_thread_free_stack() <= 0) {
	xError("x-kernel stack warning!");
	xTrace3(event, TR_ALWAYS,
		"x-kernel  stack warning [%s] for event 0x%p with thread 0x%p",
		name, self(), scout_thread_self());
	/* give up this thread: */
	scout_thread_yield();
    } /* if */
#endif /* XK_THREAD_TRACE */
} /* evCheckStack */


void
evMarkBlocked(Semaphore *s)
{
#ifdef XK_THREAD_TRACE
    Event e;

    e = self();
    if (e) {
	e->state = E_BLOCKED;
	xGetTime(&e->stopTime);
    } /* if */
#endif /* if */
} /* evMarkBlocked */


void
evMarkRunning(void)
{
#ifdef XK_THREAD_TRACE
    Event e;

    e = self();
    if (e) {
	e->state = E_RUNNING;
    } /* if */
#endif /* XK_THREAD_TRACE */
} /* evMarkRunning */

			/*** end of event.c ***/
