/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.2
 * 
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: scout_malloc.c,v $
 *
 * HISTORY
 * $Log: scout_malloc.c,v $
 * Revision 1.2  1994/06/11  00:02:31  menze
 * Added copyright notice
 *
 * Revision 1.1  1994/04/01  22:54:07  davidm
 * Initial revision
 *
 */
#include "scout.h"
#include "scout_malloc.h"
#include "scout_trace.h"
#include "scout_util.h"

/*
 * Manages memory with a doubly linked, cyclic free list.
 *	- search starts where previous search ended
 *	- contiguous blocks are coalesced on free()
 *	- free list contains memory blocks in order of increasing addresses
 */
#ifdef SCOUT_DEBUG
 /*
  * A block has magic set to MALLOC_MAGIC if it's free and
  * MALLOC_MAGIC+1 if it's allocated.
  */
# define MALLOC_MAGIC	0xC0FFEE
#endif /* SCOUT_DEBUG */

int trace_scout_malloc = 0;

struct flist {
    struct flist *next;	/* next free block */
    struct flist *prev;	/* previous free block */
    unsigned int  size;	/* size of this block (includes header size) */
#ifdef SCOUT_DEBUG
    unsigned int  magic;
#endif /* SCOUT_DEBUG */
};

static struct flist *free_list = (struct flist*) 0;

#define MINSIZE	(sizeof(struct flist) + sizeof(int))


#ifdef SCOUT_DEBUG

void
dump_free_list(void)
{
    struct flist *el = free_list;
    int free_space = 0, len = 0;
    char bogus = 0;

    printf("FL: free_list @ %p:", free_list);
    if (!el) {
        printf("FL: <EMPTY>");
    } else {
        while (el->next > el) {
            el = el->next;
        } /* while */
        do {
	    if (el->next->prev != el) {
		bogus = 1;
		printf("\nFL: BOGUS: el=%p, el->next=%p, el->next->prev=%p\n",
		       el, el->next, el->next->prev);
	    } /* if */
	    ++len;
	    free_space += el->size;
            el = el->next;
	    if (el->magic != MALLOC_MAGIC) {
		bogus = 1;
		printf("BOGUS:");
	    } /* if */
	    printf(" %p+%d", el, el->size);
        } while (el->next > el);
    } /* if */
    printf("\nFL: free space=%d\n", free_space);
    printf("FL: length of free list=%d\n", len);
    if (bogus) {
	scout_panic(__FILE__":dump_free_list: mangled free list!");
    } /* if */
} /* dump_free_list */

#endif /* SCOUT_DEBUG */


/*
 * EL is the block of memory to insert in the free list.
 * Returns the block of memory with which EL was coalesced
 * (or EL itself).
 */
static inline struct flist*
insert(struct flist *el)
{
    struct flist *p = free_list;
    register int need_insert = TRUE;

    if (!p) {
	/* list was empty: */
	el->next = el;
	el->prev = el;
	free_list = el;
    } else {
	/*
	 * search for appropriate location (the search starts potentially
	 * in the middle of the free list, that's why the conditional
	 * looks somewhat complex):
	 */
	while (!((p < el && el < p->next)
	       || (p >= p->next && (el > p || el < p->next))))
	{
	    p = p->next;
	} /* while */
	if ((char*)p + p->size == (char*)el) {
	    /* merge with predecessor block: */
#ifdef SCOUT_DEBUG
	    el->magic = 0;
#endif /* SCOUT_DEBUG */
	    p->size += el->size;
	    el = p;
	    need_insert = FALSE;
	} /* if */
	if ((char*)el + el->size == (char*)p->next) {
	    /* merge with successor block: */
#ifdef SCOUT_DEBUG
	    p->next->magic = 0;
#endif /* SCOUT_DEBUG */
	    el->size += p->next->size;
	    if (p->next == p) {
		/* take care of singleton list: */
		el->next = el;
		el->prev = el;
	    } else {
		el->next = p->next->next;
		el->next->prev = el;
		if (el != p) {
		    el->prev = p;
		    p->next = el;
		} /* if */
	    } /* if */
	    need_insert = FALSE;
	} /* if */
	if (need_insert) {
	    el->next = p->next;
	    el->prev = p;
	    p->next->prev = el;
	    p->next = el;
	} /* if */
    } /* if */
    return el;
} /* insert */


static struct flist*
more_core(int size)
{
    return 0;
} /* more_core */


void*
scout_malloc(int size)
{
    struct flist *el;
    int rsize;

    ScoutTrace(scout_malloc, TR_DETAILED, ("scout_malloc(size=%d)", size));

    /* round size up to multiple of sizeof(void*): */
    rsize = (size + sizeof(struct flist) +
	     sizeof(void*) - 1) & ~(sizeof(void*)-1);
    el = free_list;
    if (!el) {
	el = more_core(rsize);
	if (!el) {
	    return 0;
	} /* if */
    } /* if */

    /* walk free list and take first block that is large enough: */
#ifdef SCOUT_DEBUG
    if (el->magic != MALLOC_MAGIC) {
	scout_abort("scout_malloc: bad magic: %#x instead of %#x",
		    el->magic, MALLOC_MAGIC);
    } /* if */
#endif /* SCOUT_DEBUG */
    while (el->size < rsize) {
#ifdef SCOUT_DEBUG
	if (el->magic != MALLOC_MAGIC) {
	    scout_abort("xMalloc: bad magic: %#x instead of %#x",
			el->magic, MALLOC_MAGIC);
	} /* if */
#endif /* SCOUT_DEBUG */
	el = el->next;
	if (el == free_list) {
	    el = more_core(rsize);
	    if (!el) {
		return 0;
	    } /* if */
	} /* if */
    } /* while */

    if (el->size - rsize >= MINSIZE) {
	/* start next search at reduced block: */
	free_list = el;

	/* reduce block (from end): */
	el->size -= rsize;
	el = (struct flist*) ((char*)el + el->size);
	el->size = rsize;
    } else {
	/* remove block completely from free-list: */
	if (el->next == el) {
	    /* free list is empty now: */
	    free_list = 0;
	} else {
	    el->prev->next = el->next;
	    el->next->prev = el->prev;

	    /* start next search after returned block: */
	    free_list = el->next;
	} /* if */
    } /* if */

#ifdef SCOUT_DEBUG
    el->magic = MALLOC_MAGIC + 1;
#endif /* SCOUT_DEBUG */

    ScoutTrace(scout_malloc, TR_DETAILED,
	       ("scout_malloc: ptr=%p", (char*)el + sizeof(struct flist)));
    return (char*)el + sizeof(struct flist);
} /* scout_malloc */


/*
 * Allocate "size" byte such that the returned starting address
 * is a multiple of "align":
 */
void*
scout_malloc_aligned(int size, int log_alignment)
{
    struct flist *pel, *el, *nel;
    int rsize, align, tot_size, waste;
    char *aligned_addr;

    align = 1 << log_alignment;
    ScoutTrace(scout_malloc, TR_DETAILED,
	       ("scout_malloc_aligned(size=%d,align=%d)", size, align));

    /* round size up to multiple of sizeof(void*): */
    rsize = (size + sizeof(struct flist) +
	     sizeof(void*) - 1) & ~(sizeof(void*)-1);

    el = scout_malloc(rsize + align);
    pel = el - 1;
    tot_size = pel->size;
    aligned_addr = (char*)(((u_long)(el + 1) + align - 1) & ~(align - 1));
    el = (struct flist*)aligned_addr - 1;

    pel->size = (char*) el - (char*) pel;
    el->size = rsize;
#ifdef SCOUT_DEBUG
    el->magic = MALLOC_MAGIC + 1;
#endif /* SCOUT_DEBUG */
    waste = tot_size - pel->size - el->size;
    if (waste > MINSIZE) {
	nel = (struct flist*)((char*)el + rsize);
	nel->size = waste;
#ifdef SCOUT_DEBUG
	nel->magic = MALLOC_MAGIC;
#endif /* SCOUT_DEBUG */
	insert(nel);
    } else {
	el->size += waste;	/* don't bother freeing rest */
    } /* if */
    insert(pel);		/* free front space */

    return aligned_addr;
} /* scout_malloc_aligned */


void
scout_free(void *ptr)
{
    struct flist *el = (struct flist*) (ptr - sizeof(struct flist));

    ScoutTrace(scout_malloc, TR_DETAILED, ("scout_free(ptr=%p)", ptr));
#ifdef SCOUT_DEBUG
    if (el->magic != MALLOC_MAGIC + 1) {
	scout_abort("scout_free: bad magic: %#x instead of %#x",
		    el->magic, MALLOC_MAGIC + 1);
    } /* if */
    el->magic = MALLOC_MAGIC;
#endif /* SCOUT_DEBUG */

    free_list = insert(el);

    return;
} /* scout_free */


void
scout_malloc_expand (void *mem, int size)
{
    struct flist *el;

    ScoutTrace(scout_malloc, TR_EVENTS,
	       ("scout_malloc_expand: mem=%p, size=%d", mem, size));
    el = (struct flist*) mem;
    el->size = size;
#ifdef SCOUT_DEBUG
    el->magic = MALLOC_MAGIC;
#endif /* SCOUT_DEBUG */
    insert(el);
} /* scout_malloc_expand */

			/*** end of scout_malloc.c ***/
